"""
Created by Emille Ishida on 7 August 2017.

Simple function to read SNANA light curve files.

- read_snana_lc:
    Read post-SNPCC data.
"""

import numpy as np
import pandas as pd

def read_snana_lc(file_name, flux_as_DataFrame=False):
    """
    Read DES light curve in SNANA format.

    input:     file_name, str - path to light curve file
               flux_as_DataFrame, boolean - output flux as pandas DataFrame

    output:    header, dict - keywords: snid  - SN identification
                                        z     - redshift
                                        type  - SN type
                                        model - SN model
                                        pkmjd - simulated day of max
                                        sample - target or train
                                        columns - name of columns in each
                                                  filter matrix

               if flux_as_DataFrame:
                   measurements, dict - keywords:   g, r, i, z
                                        values:     np.arrays of measurements
                                                    column names in 
                                                    header['columns']
               else:
                  measurements, DataFrame - first layer: g, r, i, z
                                            second layer: header['columns']
    """

    filters=['g', 'r', 'i', 'z']

    # read light curve data
    with open(file_name, 'r') as op1:
        lin1 = op1.readlines()

    data1 = [elem.split() for elem in lin1]

    # get header data
    raw_data = dict([[line[0], line[1:]] for line in data1
                      if len(line) > 1 and line[0] != 'OBS:'])

    header = {}
    header['snid'] = int(raw_data['SNID:'][0])
    header['z'] = float(raw_data['SIM_REDSHIFT:'][0])
    header['type'] = raw_data['SIM_NON1a:']
    header['model'] = raw_data['SIM_MODEL:']
    header['pkmjd'] = float(raw_data['SIM_PEAKMJD:'][0])
    header['pkmag'] = [float(raw_data['SIM_PEAKMAG:'][j])
                       for j in range(len(filters))]
    header['columns'] = ['MJD', 'FLUXCAL', 'FLUXCALERR', 'SNR']
    header['SIM_COMMENT:'] = raw_data['SIM_COMMENT:']

    if raw_data['SNTYPE:'][0] == '-9':
        header['sample'] = 'target'
    else:
        header['sample'] = 'train'

    # get flux measurements
    flux = {}
    
    for f in filters:
        flist = []
        for line in data1:
            if len(line) > 2 and line[0] == 'OBS:' and line[2] == f:
                flist.append([float(line[1]), float(line[4]),
                              float(line[5]), float(line[6]), float(line[7]), float(line[8])])

        flux[f] = np.array(flist)

 
    if flux_as_DataFrame:

        data = {}
        for f in filters:
            data[f] = {}
            data[f]['MJD'] = flux[f][:,0]
            data[f]['FLUXCAL'] = flux[f][:,1]
            data[f]['FLUXCALERR'] = flux[f][:,2]
            data[f]['SNR'] = flux[f][:,3]
            data[f]['MAG'] = flux[f][:,4]
            data[f]['MAGERR'] = flux[f][:,5]

        data = pd.DataFrame(data)

        return header, data

    return header, flux



"""Created by Emille E. O. Ishida in 7 Sep 2016.

Collection of output functions.

- save_results
    Save results from Active Learning Pipeline to file. 

"""
import numpy as np
import pandas as pd
import os

from actsnclass import read_snana_lc


def save_results(accuracies, effs, purs, foms, queryNames, queryClasses,
                 queryLabels, user_input, path_hi_snr, path_low_snr):
    """"Save results from Active Learning Pipeline to file. 

        ***************************** IMPORTANT: *****************************
        * this is extremelly taylored for post-SNPCC data format!            *
        * if you are using another data set, writing your own print function *
        * is advisable!                                                      *
        **********************************************************************

        input: accuracies, list - accuracy results
               effs, list - efficiency results
               purs, list - purity results
               foms, list - figure of merit results
               queryNames, list - identification of queried objects
               queryClasses, list - code of template used for simulations
               queryLabels, list - sn type for each queried object
               user_input, dict - user choices
                 keywords mus contain:
                     output_folder - path to output folder
                     diagnosticFile - complete path to output diagnostics
                     queryFile - complete path to output query properties

               path_hi_snr - complete path to high SNR identification
               path_low_snr - complete path to low SNR identification
                     
        output: 2 csv files:
                   diagnostics for each new query
                   characteristics of each queried object
    """

    snIbc = ['1','5','6','7','8','9','10','11','13','14','16','18','22',
             '23','29','45','28']
    snII = ['2','3','4','12','15','17','19','20','21','24','25','26','27',
            '30','31','32','33','34','35','36','37','38','39','40','41',
            '42','43','44']

    # Read high SNR ids
    op1 = open(path_hi_snr, 'r')
    lin1 = op1.readlines()
    op1.close()

    data1 = [elem.split() for elem in lin1]
    high_SNR_ids = [int(data1[j][0][6:]) for j in range(1, len(data1))]

    # Read low SNR ids
    op2 = open(path_low_snr, 'r')
    lin2 = op2.readlines()
    op2.close()

    data2 = [elem.split() for elem in lin2]
    low_SNR_ids = [int(data2[j][0][6:]) for j in range(1, len(data2))]

    # Create number of queries vector
    indx = range(0, len(foms))

    # Create data frame with diagnostics results
    diag_rows = np.vstack((indx, accuracies, effs, purs, foms))
    diag_columns = diag_rows.transpose()
    my_df = pd.DataFrame(diag_columns, columns=['nqueries', 'accuracy', 
                                                'eff', 'pur', 'fom'])
    my_df['nqueries'] = my_df['nqueries'].astype(int)

    # Check if output folder exists
    if not os.path.isdir(user_input['output_folder']):
        os.makedirs(user_input['output_folder'])

    my_df.to_csv(user_input['output_folder'] + user_input['diagnosticFile'], 
                 index=False, header=['nqueries', 'accuracy', 'eff',
                                      'pur', 'fom'])

    query_features = []
    for i in range(len(queryNames)):

        # Read raw data
        header = read_snana_lc(rawDataPath + queryNames[i]+'.DAT')[0]

        # get SN type
        if header['type'][0] == '0':
            sntype = 'Ia'
        elif header['type'][0] in snIbc:
            sntype = 'Ibc'
        elif header['type'][0] in snII:
            sntype = 'II'
        else:
            raise ValueError('Invalid SN type!')

        # identify SNR classification
        if header['snid'] in high_SNR_ids:
            SNR = 'high'
        elif header['snid'] in low_SNR_ids:
            SNR = 'low'
        else:
            raise ValueError('snid is not listed in SNR files!')

        # identify sn type
        if queryLabels[i] == 0:
            qlabel = 'nIa'
        elif queryLabels[i] == 1:
            qlabel = 'Ia'
        else:
            raise ValueError('Invalid Query Label!')

        if qlabel == 'Ia' and sntype in ['Ibc', 'II']:
            raise ValueError('Incompatible type between raw and script labels!')
        elif qlabel == 'nIa' and sntype == 'Ia':
            raise ValueError('Incompatible type between raw and script labels!')

        # Build entry in query file
        line = [str(i + 1), str(header['snid']), sntype, SNR, str(header['z'])]
        for k in range(4):
            line.append(str(header['pkmag'][k]))
        line.append(queryClasses[i])

        query_features.append(line)

    # Build data frame
    my_df2 = pd.DataFrame(query_features)

    # save to file
    my_df2.to_csv(user_input['output_folder'] + user_input['queryFile'], 
                  index=False, header=['nqueries', 'snid', 'sntype', 'SNR',
                                       'z', 'g_pkmag', 'r_pkmag', 'i_pkmag',
                                       'z_pkmag', 'qclass'])




"""Created by CRP #4 team between 20-27 Aug 2017.

Simple function to read fitted SN light curves.

- loadData
    Load supernova data from concatenated matrix.
"""

import numpy as np

def loadData(trainDataPath, trainLabelPath, cvDataPath, cvLabelPath,
             cutFirstHalf=False, firstHalfParameters=None,
             pubDataPath=None, pubLabelPath=None):
    """
    Load supernova data from concatenated matrix.

    input: trainDataPath, str - complete path to train feature file
           trainLabelPath, str - complete path to train label file
           cvDataPath, str - complete path to cross-validation feature file
           cvLabelPath, str - complete path to cross-validation label file
           cutFirstHalf, boolean - rather to use partial light curves
                                   (optional, default is False)
           firstHalfParameters, tuple or None - column indexes for epoch cut
                                                in each filter
                                                (optional, default is None)
           trainFraction, float - fraction of training set to use
                                  (optional, default is 1) 
           pubDataPath, str - complete path to publication sample feature file
                              default is None - not to be used during analysis
           pubLabelPath, str - complete path to publication sample label file
                               default is None - not to be used during analysis

    output: tuple - trainingSet, array
                    trainingLables, list
                    crossValidationSet, array
                    crossValidationLabels, list
    """

    # Read training set
    trainingSet=np.loadtxt(trainDataPath)
    trainingLabels=np.genfromtxt(trainLabelPath, dtype="|U16")

    # Read cross-validation set
    crossValidationSet=np.loadtxt(cvDataPath)
    crossValidationLabels=np.genfromtxt(cvLabelPath, dtype="|U16")

    if pubDataPath is not None:
        # Read Publication set
        pubSet=np.loadtxt(pubDataPath)
        pubLabels=np.genfromtxt(pubLabelPath, dtype="|U16")

    else:
        pubSet = None
        pubLabels = None
    
    # Select partial light curve
    if cutFirstHalf:
    
        if (firstHalfParameters is None):
            firstHalfParameters = (4,33,11)
        
        halfCurveIndex = []
        for i in range(firstHalfParameters[0]):
            halfCurveIndex += list(range(0 + i*firstHalfParameters[1],
                                         firstHalfParameters[2] + 
                                         i * firstHalfParameters[1]))

        trainingSet = trainingSet[:,halfCurveIndex]
        crossValidationSet = crossValidationSet[:,halfCurveIndex]
        pubSet = pubSet[:,halfCurveIndex]
    
    return (trainingSet,trainingLabels,crossValidationSet,
                  crossValidationLabels, pubSet, pubLabels)



from snactclass.lc_functions.light_curve_functions import sanders, karpenka

import matplotlib.pyplot as plt
import numpy as np



def plot_karpenka(trace,flux,burnin=1000,thin=20,alpha=.09):
    
   
    fig, ax = plt.subplots()

    for point in trace[burnin::thin]:

        for band, color in zip(['g','r','i','z'],["#dd0100","#fac901","#225095",'#9b4800']): 
            
            err = np.array(flux[band]['FLUXCALERR'])
            idx = np.isfinite(err)
            time = np.array(flux[band]['MJD'] - flux[band]['MJD'].min())[idx]
            
            pseudo_time = np.linspace(time.min(),time.max(),50)

            pred = karpenka(pseudo_time,
                          10**point['logA_%s'%band],
                          10**point['logB_%s'%band],
                          point['t0_%s'%band],
                          point['t1_%s'%band],
                          point['t_fall_%s'%band],
                          point['t_rise_%s'%band]
                         )


            ax.plot(psudo_time,pred,color,alpha=.09,zorder=-33)


    for band, color in zip(['g','r','i','z'],["#dd0100","#fac901","#225095",'#9b4800']):    


        err = np.array(flux[band]['FLUXCALERR'])
        idx = np.isfinite(err)
        err = err[idx]
        tdata = np.array(flux[band]['FLUXCAL'])[idx]
        time = np.array(flux[band]['MJD'] - flux[band]['MJD'].min())[idx]


        ax.errorbar(time,tdata,yerr=err,color=color,fmt='o',zorder=-1,label='%s data'%band)
        
        
    ax.set_xlabel('time')
    ax.set_ylabel('flux')
    ax.legend()
    
    return fig



def plot_sanders(trace, flux, burnin=1000,thin=20,alpha=.09):
    
   
    fig, ax = plt.subplots()

    for point in trace[burnin::thin]:

        for band, color in zip(['g','r','i','z'],["#dd0100","#fac901","#225095",'#9b4800']): 
            
            err = np.array(flux[band]['FLUXCALERR'])
            idx = np.isfinite(err)
            time = np.array(flux[band]['MJD'] - flux[band]['MJD'].min())[idx]
            
            pseudo_time = np.linspace(time.min(),time.max(),50)

            pred = sanders(pseudo_time,
                  10**point['logYb_%s'%band],
                  10**point['logMp_%s'%band],
                  10**point['logalpha_%s'%band],
                  10**point['logbeta1_%s'%band],
                  10**point['logbeta2_%s'%band],
                  10**point['logbetadn_%s'%band],
                  10**point['logbetadc_%s'%band],
                  point['t0_%s'%band],
                  point['t1_%s'%band],
                  point['tp_%s'%band],
                  point['t2_%s'%band],
                  point['td_%s'%band]
                 )


            ax.plot(psudo_time,pred,color,alpha=.09,zorder=-33)


    for band, color in zip(['g','r','i','z'],["#dd0100","#fac901","#225095",'#9b4800']):    


        err = np.array(flux[band]['FLUXCALERR'])
        idx = np.isfinite(err)
        err = err[idx]
        tdata = np.array(flux[band]['FLUXCAL'])[idx]
        time = np.array(flux[band]['MJD'] - flux[band]['MJD'].min())[idx]


        ax.errorbar(time,tdata,yerr=err,color=color,fmt='o',zorder=-1,label='%s data'%band)
        
        
    ax.set_xlabel('time')
    ax.set_ylabel('flux')
    ax.legend()
    
    return fig



"""Created by the CRP #4 team between 20-27 Aug 2017.

Functions to deal with the naming conventions on post-SNPCC data.

- plt_lc_raw
    Generate plot with the observed points in a given light curve. 
"""


import pandas as pd
import matplotlib.pyplot as plt

def plot_lc_raw(lc):
    """
    Generate plot with the observed points in a given light curve. 

    input: lc, tuple - output from snactclass.read_SNANA.read_snana_lc 

    output: plot visualization
    """
    
    filters = ['g', 'r', 'i', 'z']

    data = {}
    for f in filters:
        data[f] = {}
        data[f]['MJD'] = lc[1][f][:,0]
        data[f]['FLUXCAL'] = lc[1][f][:,1]
        data[f]['FLUXCALERR'] = lc[1][f][:,2]

    data = pd.DataFrame(data)
    
    fig, ax = plt.subplots()

    color_wheel = {'r':'#dd0100','g':'#fac901','i':'#225095','z':'brown'}
    for i, (band, data) in enumerate(data.iteritems()):
        
        ax.errorbar(data['MJD'],
                    data['FLUXCAL'],
                    yerr=data['FLUXCALERR'],
                    color=color_wheel[band],
                    fmt='o',
                    label=band)
        
    ax.legend()
    ax.set_xlabel('MJD')
    ax.set_ylabel('Flux')

    plt.show()
    



import numpy as np

from actsnclass.analysis_functions.diagnostics import efficiency, purity, fom

def learnLoop(quota, trainData, fullLabels, validationFeatures,
              validationClasses, labeler, mainModelList, queryStrategy,
              queryStrategyID):
    
    accuracies=[]
    efficiencies=[]
    purities=[]
    foms=[]
    
    queryNames=[]
    queryClasses=[]
    queryLabels=[]
    
    #Could optimise this based on queryStrategyID, but not vital since it only runs once
    [model.train(trainData) for model in mainModelList]
    
    predictedClasses = np.array([model.predict(validationFeatures) for model in mainModelList])
    predictedClasses = np.array(np.sum(predictedClasses, axis=0) > len(mainModelList)/2.0, dtype=np.int32)
    
    accuracies.append(np.mean(predictedClasses == validationClasses))
    efficiencies.append(efficiency(predictedClasses,validationClasses,Ia_flag=1))
    purities.append(purity(predictedClasses,validationClasses,Ia_flag=1))
    foms.append(fom(predictedClasses,validationClasses,Ia_flag=1))
    
    
    for i in range(quota):
        # Standard usage of libact objects
        if i % 5 == 0:
            print('Queries '+str(i))
  
        #UncertaintySampling fits the models here, inefficient
        query_ids = queryStrategy.make_query()
        
        if (queryStrategyID==3 or queryStrategyID==4 or queryStrategyID==5):
            #These are the batch query strategies
            
            labels=[]
            for query_id in query_ids:
                labels.append(labeler.label(trainData.data[query_id][0]))
            
                queryNames.append(fullLabels[query_id,0])
                queryClasses.append(fullLabels[query_id,2])
                queryLabels.append(labels[-1])

            queryStrategy.update_batch(query_ids, labels)  
            #The update_batch fits the models for N-least certain and semisupervised
            #Random does not have an associated model

        else:            
            label=labeler.label(trainData.data[query_ids][0])
            
            #The dataset update fits the models for QueryByCommittee via a callback function
            #Random does not have an associated model, UncertaintySampling isn't updated
            trainData.update(query_ids, label)
        
            queryNames.append(fullLabels[query_ids,0])
            queryClasses.append(fullLabels[query_ids,2])
            queryLabels.append(label)            
            
            #queryStrategy.update(query_id, label)
        
        
        #UncertaintySampling needs to be fitted here with the updated data, inefficient as it is also fitted in the make_query
        #Query by committee might have a separate main model, in that case it has to be fitted too
        #Random methods need the model fitted too
        if (queryStrategyID==0 or queryStrategyID==1 or queryStrategyID==3 
            or (queryStrategyID==2 and len(mainModelList)<queryStrategy.n_students)):

            [model.train(trainData) for model in mainModelList]
        
        predictedClasses = np.array([model.predict(validationFeatures) for model in mainModelList])
        predictedClasses = np.array(np.sum(predictedClasses, axis=0) > len(mainModelList)/2.0, dtype=np.int32)
        
        accuracies.append(np.mean(predictedClasses == validationClasses))
        efficiencies.append(efficiency(predictedClasses,validationClasses,Ia_flag=1))
        purities.append(purity(predictedClasses,validationClasses,Ia_flag=1))
        foms.append(fom(predictedClasses,validationClasses,Ia_flag=1))

    return (accuracies, efficiencies, purities, foms, queryNames, queryClasses, queryLabels, predictedClasses)



# Copyright 2020 snactclass software
# Author: Emille E. O. Ishida
#         Based on initial prototype developed by the CRP #4 team
#
# created on 26 February 2020
#
# Licensed GNU General Public License v3.0;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.gnu.org/licenses/gpl-3.0.en.html
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from copy import deepcopy
import numpy as np
import pandas as pd

from actsnclass import LightCurve


class PLAsTiCCPhotometry(object):
    """Handles photometric information for the PLAsTiCC data.

    Attributes
    ----------
    bazin_header: str
        Header to be added to features files for each day.
    class_code: dict
        Keywords are model numbers, values are model names. 
    fdic: dict
        Keywords: ['test', 'train'], values are names of photo files.
    include_classes: list
        List of classes to be considered.
    max_epoch: float
        Maximum MJD for the entire data set.
    metadata: dict
        Keywords: ['test', 'train'], values are metadata.
    min_epoch: float
        Minimum MJD for the entire data set.
    rmag_lim: float
        Maximum r-band magnitude allowing a query. 
    
    Methods
    -------
    build_one_epoch(raw_data_dir: str, day_of_survey: int,
                    time_domain_dir: str, feature_method: str)
        Construct 1 entire day of observation.
    create_all_daily_files(raw_data_dir: str)
        Create 1 file per day for all days of the survey. 
    create_daily_file(output_dir: str, day: int, vol: int, header: str)
        Create one file for a given day of the survey. Contains only header.    
    fit_one_lc(raw_data_dir: str, snid: int, sample: str)
        Fit one light curve throughout the entire survey.
    read_metadata(path_to_data_dir: str, classes: list)
        Read metadata and filter only required classes.
    write_bazin_to_file(lightcurve: LightCurve, features_file: str)
        Write Bazin parameters and metadata to file.
    """

    def __init__(self):
        self.bazin_header = 'id redshift type code sample queryable uA uB ut0 ' + \
                            'utfall utrise gA gB gt0 ' + \
                            'gtfall gtrise rA rB rt0 rtfall rtrise iA ' + \
                            'iB it0 itfall itrise zA zB zt0 ztfall ztrise' + \
                            'YA YB Yt0 Ytfall Ytrise\n'
        self.class_code = {90: 'Ia', 67: '91bg', 52:'Iax', 42:'II', 62:'Ibc', 
             95: 'SLSN', 15:'TDE', 64:'KN', 88:'AGN', 92:'RRL', 65:'M-dwarf',
             16:'EB',53:'Mira', 6:'MicroL', 991:'MicroLB', 992:'ILOT', 
             993:'CART', 994:'PISN',995:'MLString'}
        self.fdic = {}
        self.fdic['test'] = ['plasticc_test_lightcurves_' + str(x).zfill(2) + '.csv.gz' 
                              for x in range(1, 12)]
        self.fdic['train'] = ['plasticc_train_lightcurves.csv.gz']
        self.include_classes = [] 
        self.max_epoch = 60675
        self.metadata = {}
        self.min_epoch = 59580
        self.rmag_lim = 24
        
    def create_daily_file(self, output_dir: str,
                          day: int, vol: int, header='Bazin'):
        """Create one file for a given day of the survey.

        The file contains only header for the features file.

        Parameters
        ----------
        output_dir: str
            Complete path to raw data directory.
        day: int
            Day passed since the beginning of the survey.
        vol: int
            Index to identify when there is more than 1 file per day.
        header: str (optional)
            List of elements to be added to the header.
            Separate by 1 space.
            Default option uses header for Bazin features file.
        """

        features_file = output_dir + 'day_' + str(day) + '_v' + str(vol) +'.dat'

        if header == 'Bazin':
            # add headers to files
            with open(features_file, 'w') as param_file:
                param_file.write(self.bazin_header)

        else:
            with open(features_file, 'w') as param_file:
                param_file.write(self.header)


    def read_metadata(self, path_to_data_dir: str, classes: list):
        """ Read metadata and filter only required classes.

        Populate the metadata attribute. 

        Parameters
        ----------
        path_to_data_dir: str
            Directory containing all PlAsTiCC files. 
        classes: list of int
            Codes for classes we wish to keep. 

        """
        flist = ['plasticc_train_metadata.csv.gz', 
                  'plasticc_test_metadata.csv.gz']

        # store classes information
        self.include_classes = classes

        # read header information
        for i in range(2):
            if '.tar.gz' in flist[i]:
                tar = tarfile.open(path_to_data_dir + flist[i], 'r:gz')
                name = tar.getmembers()[0]
                content = tar.extracfile(name).read()
                header = pd.read_csv(io.BytesIO(content))
                tar.close()
            else:
                header = pd.read_csv(path_to_data_dir + flist[i], 
                                     index_col=False)
                if ' ' in header.keys()[0]:
                    header = pd.read_csv(path_to_data_dir + flist[i], 
                                         sep=' ', index_col=False)
            
            # remove useless columns
            header = header[['object_id', 'true_target', 'true_z',
                             'true_peakmjd', 'true_distmod']]

            class_flag = np.array([header['true_target'].iloc[j] in classes
                                   for j in range(header.shape[0])])

            if i == 0:
                self.metadata['train'] = header[class_flag]
            else:
                self.metadata['test'] = header[class_flag]

    def create_all_daily_files(self, raw_data_dir: str, output_dir:str):
        """Create 1 file per day for all days of the survey. 
        
        Each file contains only the header. 

        Parameters
        ----------
        raw_data_dir: str
            Path to directory containing all PLAsTiCC zenodo data.
        output_dir: str
            Output directory.
        """
        
        ## create daily files
        # run through training and test
        for key in ['train', 'test']:

            # run through multiple photometric files
            for j in range(len(self.fdic[key])):

                # identify number of photometric test files
                if key == 'train':
                    vol = j
                else:
                    vol = j + 1

                # run through all days of the survey
                for day_of_survey in range(1, self.max_epoch - self.min_epoch):
                    self.create_daily_file(output_dir=output_dir,
                                           day=day_of_survey, vol=vol)

    def write_bazin_to_file(self, lc: LightCurve, 
                            features_file: str, queryable: bool):
        """Write Bazin parameters and metadata to file.

        Use output filename defined in the features_file attribute.

        Parameters
        ----------
        lc: LightCurve
            ActSNClass light curve object.
        features_file: str
            Output file to store Bazin features.
        queryable: bool
            Rather this object is available for querying.
        
        Returns
        -------
        line: str
            A line concatenating metadata and Bazin fits for 1 obj.
        """

        # build an entire line with bazin features
        line = str(lc.id) + ' ' + str(lc.redshift) + ' ' + \
               str(lc.sntype) + ' ' + str(lc.sncode) + ' ' + \
               str(lc.sample) + ' ' + str(queryable) + ' '

        for item in lc.bazin_features[:-1]:
            line = line + str(item) + ' '

        line = line + str(lc.bazin_features[-1]) + '\n'
        
        # save features to file
        with open(features_file, 'a') as param_file:
            param_file.write(line)

        return line
            
    def fit_one_lc(self, raw_data_dir: str, snid: int, sample: str,
                   output_dir: str):
        """Fit one light curve throughout the entire survey.

        Save results to appropriate file, considering 1 day survey 
        evolution. 

        Parameters
        ----------
        raw_data_dir: str
            Complete path to all PLAsTiCC zenodo files.
        snid: int
            Object id for the transient to be fitted.
        sample: str
            Original sample this object belonged to.
            Possibilities are 'train' or 'test'.
        output_dir:
            Directory to store output time domain files.
        """

        # store number of points per day
        npoints = {}
        npoints[0] = 0

        # create light curve instance
        orig_lc = LightCurve()          

        vol = 0
        if sample == 'train':
            # load light curve
            orig_lc.load_plasticc_lc(raw_data_dir + self.fdic[sample][vol], snid)

        else:
            # search within test light curve files
            while orig_lc.photometry.shape[0] == 0:
                vol = vol + 1
                orig_lc.load_plasticc_lc(raw_data_dir + self.fdic[sample][vol], snid)
        
        line = False
                
        # for every day of survey
        for day_of_survey in range(1, self.max_epoch - self.min_epoch):

            lc = deepcopy(orig_lc)

            # see which epochs are observed until this day
            today = day_of_survey + self.min_epoch
            photo_flag = lc.photometry['mjd'].values <= today

            # number of points today
            npoints[day_of_survey] = sum(photo_flag)

            # only the allowed photometry
            lc.photometry = lc.photometry[photo_flag]

            # check if any point survived, other checks are made
            # inside lc object
            if npoints[day_of_survey] > 4 and \
               npoints[day_of_survey] - npoints[day_of_survey - 1] > 0:

                # perform feature extraction
                lc.fit_bazin_all()
   
                # only save to file if all filters were fitted
                if len(lc.bazin_features) > 0 and \
                        'None' not in lc.bazin_features:

                    # see if query is possible
                    queryable = \
                        lc.check_queryable(mjd=self.min_epoch + day_of_survey,
                                           r_lim=self.rmag_lim)              

                    # mask only metadata for this object
                    mask = self.metadata[sample]['object_id'].values == snid
                    
                    # set redshift
                    lc.redshift = self.metadata[sample]['true_z'].values[mask][0]

                    # set light curve type
                    lc.sncode  = self.metadata[sample]['true_target'].values[mask][0]
                    lc.sntype = self.class_code[lc.sncode]

                    # set id
                    lc.id = snid

                    # set sample
                    lc.sample = sample

                    # set filename
                    features_file = output_dir + 'day_' + \
                                         str(day_of_survey) + '_v' + str(vol) +'.dat'

                    print(day_of_survey)
                    # write to file
                    line = self.write_bazin_to_file(lc, features_file, queryable)
                    
            # write previous result to file if there is no change
            elif npoints[day_of_survey] == npoints[day_of_survey - 1] and isinstance(line, str):

                # set filename
                features_file = output_dir + 'day_' + \
                                str(day_of_survey) + '_v' + str(vol) +'.dat'

                print(day_of_survey, '****')
                # save results
                with open(features_file, 'a') as param_file:
                    param_file.write(line)
                
    def build_one_epoch(self, raw_data_dir: str, day_of_survey: int,
                        time_domain_dir: str, feature_method='Bazin'):
        """Construct 1 entire day of observation. 

        Save results to file.

        Parameters
        ----------
        raw_data_dir: str
            Directory containing all the PLAsTiCC zenodo data.
        day_of_survey: int
            Day since the beginning of the survey to be constructed.
        time_domain_dir: str
            Output directory
        feature_method: str
            Feature extraction method. 
            Only possibility now is 'Bazin'.
        """
        
        # count survivers
        count_surv = 0

        # run through training and test
        for key in ['train', 'test']:

            # run through multiple photometric files
            for j in range(len(self.fdic[key])):

                # create features file
                if key == 'train':
                    vol = j
                else:
                    vol = j + 1
                    
                self.create_daily_file(output_dir=time_domain_dir,
                                       day=day_of_survey, vol=vol)

                features_file = time_domain_dir + 'day_' + str(day_of_survey) + \
                                '_v' + str(vol) +'.dat'
                
                for i in range(self.metadata[key].shape[0]):

                    # choose 1 snid
                    snid =  self.metadata[key]['object_id'].iloc[i]
 
                    lc = LightCurve()  # create light curve instance

                    # define original sample
                    lc.sample = key
  
                    if dataset == 'PLAsTiCC':
                        lc.load_plasticc_lc(raw_data_dir + self.fdic[key][j], snid)
                    else:
                        raise ValueError('Only PLAsTiCC data set is ' + 
                                         'implemented in this module!')

                    # see which epochs are observed until this day
                    today = day_of_survey + self.min_epoch
                    photo_flag = lc.photometry['mjd'].values <= today

                    # check if any point survived, other checks are made
                    # inside lc object
                    if sum(photo_flag) > 4:
                        # only the allowed photometry
                        lc.photometry = lc.photometry[photo_flag]

                        # perform feature extraction
                        if feature_method == 'Bazin':
                            lc.fit_bazin_all()
                        else:
                            raise ValueError('Only Bazin features are implemented!')
   
                        # only save to file if all filters were fitted
                        if len(lc.bazin_features) > 0 and \
                                'None' not in lc.bazin_features:

                            # see if query is possible
                            queryable = \
                                lc.check_queryable(mjd=self.min_epoch + day_of_survey,
                                                   r_lim=self.rmag_lim)              

                            # set redshift
                            lc.redshift = self.metadata[key]['true_z'].iloc[i]

                            # set light curve type
                            lc.sncode  = self.metadata[key]['true_target'].iloc[i]
                            lc.sntype = self.class_code[lc.sncode]

                            # set id
                            lc.id = snid
                            
                            # save results
                            self.write_bazin_to_file(lc, features_file, queryable)
    



"""Created by Emille E. O. Ishida in 7 Sep 2016.

Collection of output functions.

- save_results
    Save results from Active Learning Pipeline to file. 

"""
import numpy as np
import pandas as pd
import os

from actsnclass.data_functions.read_SNANA import read_snana_lc
from actsnclass.analysis_functions.ml_result import MLResult


def save_results(accuracies, effs, purs, foms, queryNames, queryClasses,
                 queryLabels, user_input, path_raw_data, Ia_label=1, path_features=None):
    """"Save results from Active Learning Pipeline to file. 

        ***************************** IMPORTANT: *****************************
        * this is extremelly taylored for post-SNPCC data format!            *
        * if you are using another data set, writing your own print function *
        * is advisable!                                                      *
        **********************************************************************

        input: accuracies, list - accuracy results
               effs, list - efficiency results
               purs, list - purity results
               foms, list - figure of merit results
               queryNames, list - identification of queried objects
               queryClasses, list - code of template used for simulations
               queryLabels, list - sn type for each queried object
               user_input, dict - user choices
                 keywords mus contain:
                     outputFolder - path to output folder
                     diagnosticFile - complete path to output diagnostics
                     queryFile - complete path to output query properties
               path_row_data, str - path to raw data folder
                     
        output: 2 csv files:
                   diagnostics for each new query
                   characteristics of each queried object
    """

    snIbc = ['1','5','6','7','8','9','10','11','13','14','16','18','22',
             '23','29','45','28']
    snII = ['2','3','4','12','15','17','19','20','21','24','25','26','27',
            '30','31','32','33','34','35','36','37','38','39','40','41',
            '42','43','44']

    filters = ['g', 'r', 'i', 'z']

    # Check if output folder exists
    if not os.path.isdir(user_input['outputFolder']):
        os.makedirs(user_input['outputFolder'])
    
    

    # save diagnostics results
    if user_input['choice'] in ['batch_random', 'batch_nlunc', 'batch_semi']:

        # horrible way to take into account the many ids in a batch
        # (repeating the diagnostics)
        [queryNames.insert(0, '---') for k in range(user_input['batchSize'])]
        foms_rep = [foms[i] for i in range(len(foms)) \
                            for j in range(user_input['batchSize'])]
        effs_rep = [effs[i] for i in range(len(effs)) \
                            for j in range(user_input['batchSize'])]
        purs_rep = [purs[i] for i in range(len(purs)) \
                            for j in range(user_input['batchSize'])]
        accuracies_rep = [accuracies[i] for i in range(len(foms)) \
                                     for j in range(user_input['batchSize'])]

        diagnostics = MLResult(queryNames, foms_rep, effs_rep, purs_rep, accuracies_rep)
            
    else:
        queryNames.insert(0, '---')
        diagnostics = MLResult(queryNames, foms, effs, purs, accuracies)
    
    # save to file
    diagnostics.save(user_input['outputFolder'] + user_input['diagnosticFile'])

    if user_input['choice'] in ['batch_random', 'batch_nlunc', 'batch_semi']:
        [queryNames.remove('---') for k in range(user_input['batchSize'])]
    else:
        queryNames.remove('---')

    query_features = []
    for i in range(len(queryNames)):

        # Read raw data
        lc = read_snana_lc(path_raw_data + queryNames[i]+'.DAT')

        # get SN type
        if lc[0]['type'][0] == '0':
            sntype = 'Ia'
        elif lc[0]['type'][0] in snIbc:
            sntype = 'Ibc'
        elif lc[0]['type'][0] in snII:
            sntype = 'II'
        else:
            raise ValueError('Invalid SN type!')

        # identify sn type
        if queryLabels[i] != Ia_label:
            qlabel = 'nIa'
        elif queryLabels[i] == Ia_label:
            qlabel = 'Ia'
        else:
            raise ValueError('Invalid Query Label!')

        if qlabel == 'Ia' and sntype in ['Ibc', 'II']:
            raise ValueError('Incompatible type between raw and script labels!')
        elif qlabel == 'nIa' and sntype == 'Ia':
            raise ValueError('Incompatible type between raw and script labels!')

        # Build entry in query file
        # line = [str(i + 1), str(header['snid']), sntype, SNR, str(header['z'])]
        line = [str(i + 1), str(lc[0]['snid']), sntype, str(lc[0]['z'])]
        for k in range(4):
            line.append(str(lc[0]['pkmag'][k]))

        for f in filters:
            line.append(str(np.mean(lc[1][f][:,3])))        

        line.append(queryClasses[i])

        query_features.append(line)

    # Build data frame
    my_df2 = pd.DataFrame(query_features)
 
    h1 = ['nqueries', 'snid', 'sntype', 'z', 'g_pkmag', 'r_pkmag',
          'i_pkmag', 'z_pkmag', 'g_SNR', 'r_SNR', 'i_SNR', 'z_SNR', 'qclass']
 
    # save to file
    my_df2.to_csv(user_input['outputFolder'] + user_input['queryFile'], 
                  index=False, header=h1)




from light_curve_functions import bazin, karpenka, sanders
from light_curve_functions import newling
from light_curve_plot import plot_lc_raw

from pymc_plotting import plot_karpenka, plot_sanders

from sn_observation import SNObservation, SNBand, FittedObservation



"""Created by CRP #4 team between 20-27 Aug 2017.

Simple function to randomize objects in a given sample.

- randomiseData
    Separates the complete cross-validation set into query and validation.
"""

import numpy as np

def randomiseData(trainFeatures, trainLabels, cvFeatures,
                  cvLabels, queryFraction=0.8,
                  fixRandomState=False, trainFraction=1.0,
                  pubFeatures=None, pubLabels=None):
    """Separates the complete target set into query and validation sets, 
       with the appropriate randomization.

    input: trainFeatures, array - complete training sample features
           trainLabels, list - complete training sample lables
           cvFeatures, array - complete cross-validation sample features
           cvLabels, list - complete cross-validation sample labels
           queryFraction, float - fraction of cv sample to be used to query
                                  (optional, default is 0.8)
           fixRandomState, boolean - rather to use a fixed random seed
                                     (optional, default is False)
           trainFraction, float - fraction of training set to be used
                                  (optional, default is 1.0)
           pubFeatures - complete publication sample features
                            (this is only used if queryFraction == 0,
                             this correspond to publication option when 
                             cv sample is static, default is None)
           pubLabeles - complete publication sample labels
                            (this is only used if queryFraction == 0,
                             this correspond to publication option when 
                             cv sample is static, default is None)

    output: data, dict - separated train, query and validation sets
              keywords: trainFeatures, trainLabels, 
                        queryPoolFeatures, queryPoolLabels, 
                        validFeatures, validLabels
    """
    
    if (fixRandomState):
        np.random.seed(42)
    
    # Count the number of objects in each set
    nTest = len(cvLabels)
    nTrain = len(trainLabels)

    # Get indexes for test set
    inds = np.arange(nTest)

    # Randomize indexes for test set
    np.random.shuffle(inds)

    if queryFraction > 0:
        # Separate query set indexes
        testInds = inds[:int(nTest * queryFraction)]

        # Separate validation set indexes
        validInds = inds[int(nTest * queryFraction):]

        # Get query and validation data
        queryPoolFeatures, queryPoolLabels = cvFeatures[testInds], cvLabels[testInds]
        validFeatures, validLabels = cvFeatures[validInds], cvLabels[validInds]

    else:
        # complete query is given by the cv sample
        # complete validation is given by the publication sample
        queryPoolFeatures, queryPoolLabels = cvFeatures, cvLabels
        validFeatures, validLabels = pubFeatures, pubLabels
        

    # Now shuffle up the trainingInputs too
    inds = np.arange(nTrain)
    np.random.shuffle(inds)

    # Separate the useful trainig set indexes
    trainInds = np.array([inds[i] for i in range(int(nTrain * trainFraction))])

    # Get training data
    trainFeatures = trainFeatures[trainInds]
    trainLabels = trainLabels[trainInds]

    # Organize output
    data = {}
    data['trainFeatures'] = trainFeatures
    data['trainLabels'] = trainLabels
    data['queryPoolFeatures'] = queryPoolFeatures
    data['queryPoolLabels'] = queryPoolLabels
    data['validFeatures'] = validFeatures
    data['validLabels'] = validLabels

    return data




"""Created by CRP #4 team between 20-27 Aug 2017

Standard diagnostic to access classification results. 

- efficiency
    Fraction of type Ia SNe found.
- purity
    Fraction of correct Ia classifications.
- fom
    efficiency * pseudo-purity.
"""

def efficiency(label_pred, label_true, Ia_flag=0):
    """Calculate efficiency.
       
       input: label_pred, predicted labels
              label_true, true labels
              Ia_flag, Ia flag (optional, default is 0)

       output: efficiency
    
    """

    cc_Ia = sum([label_pred[i] == label_true[i] and label_true[i] == Ia_flag for i in range(len(label_pred))])
    tot_Ia = sum([label_true[i] == Ia_flag for i in range(len(label_true))])

    return float(cc_Ia)/tot_Ia


def purity(label_pred, label_true, Ia_flag=0):
    """Calculate purity.

       input: label_pred, predicted labels
              label_true, true labels
              Ia_flag, Ia flag (optional, default is 0)

       output: purity
    """

    cc_Ia = sum([label_pred[i] == label_true[i] and label_true[i] == Ia_flag for i in range(len(label_pred))])
    wr_nIa = sum([label_pred[i] != label_true[i] and label_true[i] != Ia_flag for i in range(len(label_pred))])

    if cc_Ia + wr_nIa > 0:
        return float(cc_Ia)/(cc_Ia + wr_nIa)
    else:
        return 0

def fom(label_pred, label_true, Ia_flag=0, penalty=3):
    """
    Calculate figure of merit.

    input: label_pred, predicted labels
           label_true, true labels
           Ia_flag, Ia flag (optional, default is 0)
           penalty factor for wrong non-Ia classification (optional, default is 3)

    output: figure of merity
    """

    cc_Ia = sum([label_pred[i] == label_true[i] and label_true[i] == Ia_flag for i in range(len(label_pred))])
    wr_nIa = sum([label_pred[i] != label_true[i] and label_true[i] != Ia_flag for i in range(len(label_pred))])
    tot_Ia = sum([label_true[i] == Ia_flag for i in range(len(label_true))])

    if (cc_Ia + penalty * wr_nIa) > 0:
        return (float(cc_Ia)/(cc_Ia + penalty * wr_nIa)) * float(cc_Ia)/tot_Ia
    else:
        return 0



"""Created by CRP #4 team between 20-27 Aug 2017.

Functions to setup configuration on active learning routines.

- initialDataSetup
    Load supernova data from concatenated matrix.
"""

import copy
import numpy as np
import libact

from libact.labelers import IdealLabeler
from libact.query_strategies import RandomSampling,QueryByCommittee
from libact.query_strategies import UncertaintySampling
from libact.base.dataset import Dataset
from libact.models import SklearnAdapter, SklearnProbaAdapter

from libact.models import LogisticRegression
from sklearn.ensemble import RandomForestClassifier
from sklearn.svm import SVC
from sklearn.tree import DecisionTreeClassifier
from sklearn.neighbors import KNeighborsClassifier

from .batchQuery import RandomBatchQuery,LeastCertainBatchQuery
from .batchQuery import SemiSupervisedBatchQuery

def initialDataSetup(trainFeatures, trainClasses, testFeatures,
                     testClasses, SNLabel='0'):
    """
    Set up ideal labeler.

    input: trainFeatures, array - train matrix
           trainClasses, list - train labels
           testFeatures, array - test (photometric) matrix
           testClasses, list - test (photometric) labels
           SNLabel, str - SN Ia flag

    output: tuple, (train_dataset, fullLabels, labeler)
    """

    # Concatenate features
    fullFeatures = np.vstack([trainFeatures, testFeatures])

    # Include None in place of labels from target sample
    partialClasses = np.concatenate([(trainClasses[:,2]==SNLabel).astype(int),
                                      np.array([None] * testFeatures.shape[0])])

    # Complete concatenated labels for train and target samples
    fullClasses = np.concatenate([(trainClasses[:,2] == SNLabel).astype(int),
                                  (testClasses[:,2] == SNLabel).astype(int)])

    # Concatenate labels
    fullLabels = np.concatenate([trainClasses, testClasses])

    # Concatenated features and class labels with None on target data
    train_dataset = Dataset(fullFeatures, partialClasses)
    
    # Define ideal labeler
    labeler = IdealLabeler(Dataset(fullFeatures, fullClasses))

    return (train_dataset, fullLabels, labeler)

    
def initialModelSetup(modelID, modelParams=None, fixRandomState=False):
    
    if modelID==0:
        # Single random forest
        if modelParams is None:
            modelParams=(1000,)
        model=[SklearnProbaAdapter(RandomForestClassifier(random_state=42 \
                   if fixRandomState else None, n_estimators=modelParams[0]))]

    elif modelID==1:
        # List of random forests
        if modelParams is None:
            modelParams=(100,15)
        
        model=[SklearnProbaAdapter(RandomForestClassifier(random_state=i \
                   if fixRandomState else None, n_estimators=modelParams[0]))
                                              for i in range(modelParams[1])]

    elif modelID==2:
        # Small varied committee
        
        # SVC can be made a probabilistic model with probability=true,
        # but that slows down the fit
        
        model=[LogisticRegression(C=1.0, random_state=0 \
                                  if fixRandomState else None),
               SklearnProbaAdapter(RandomForestClassifier(n_estimators=100,
                                random_state=1 if fixRandomState else None)),
               SklearnAdapter(SVC(random_state=2 \
                                  if fixRandomState else None)),
               SklearnProbaAdapter(DecisionTreeClassifier(random_state=3 \
                                               if fixRandomState else None)),
               SklearnProbaAdapter(KNeighborsClassifier(n_neighbors=19))]
                     
    elif modelID==3:
        # Large varied committee
        
        # SVC can be made a probabilistic model with probability=true,
        # but that slows down the fit

        model=[LogisticRegression(C=1.0, 
                                  random_state=0 if fixRandomState else None),
               LogisticRegression(C=0.1, random_state=1 \
                                                 if fixRandomState else None),
               SklearnProbaAdapter(RandomForestClassifier(n_estimators=100,
                                 random_state=2 if fixRandomState else None)),
               SklearnProbaAdapter(RandomForestClassifier(n_estimators=200,
                                 random_state=3 if fixRandomState else None)),
               SklearnProbaAdapter(RandomForestClassifier(n_estimators=300,
                                 random_state=4 if fixRandomState else None)),
               SklearnProbaAdapter(RandomForestClassifier(n_estimators=400,
                                 random_state=5 if fixRandomState else None)),
               SklearnProbaAdapter(RandomForestClassifier(n_estimators=500,
                                 random_state=6 if fixRandomState else None)),
               SklearnAdapter(SVC(C=1.0,#kernel='rbf',
                                 random_state=7 if fixRandomState else None)),
               SklearnAdapter(SVC(C=0.1,kernel='rbf',
                                 random_state=8 if fixRandomState else None)),
               SklearnAdapter(SVC(C=1.0,kernel='poly', degree=3,
                                random_state=9 if fixRandomState else None)),
               SklearnAdapter(SVC(C=0.1,kernel='poly', degree=3,
                               random_state=10 if fixRandomState else None)),
               SklearnProbaAdapter(DecisionTreeClassifier(random_state=11 
                                               if fixRandomState else None)),
               SklearnProbaAdapter(KNeighborsClassifier(n_neighbors=19, 
                                                        weights='distance')),
               SklearnProbaAdapter(KNeighborsClassifier(n_neighbors=19)),
               SklearnProbaAdapter(KNeighborsClassifier(n_neighbors=9))
               ]

    return model
    
def initialQuerySetup(train_dataset, queryStrategyID,
                      queryParams=None, fixRandomState=False):

    if queryStrategyID==0:
        queryStrategy = RandomSampling(train_dataset,random_state=137 \
                                       if fixRandomState else None)

    elif queryStrategyID==1:
        queryStrategy = UncertaintySampling(train_dataset,
                                 method='sm', model=queryParams[0])

    elif queryStrategyID==2:    
        queryStrategy = QueryByCommittee(train_dataset,
                                         models=queryParams[0],
                                         disagreement='vote',
                                         random_state=23 \
                                         if fixRandomState else None)              
    elif queryStrategyID==3:
        queryStrategy = RandomBatchQuery(train_dataset,
                                         batch_size=queryParams[0],
                                         random_state=2311 \
                                         if fixRandomState else None) 

    elif queryStrategyID==4:
        queryStrategy = LeastCertainBatchQuery(train_dataset,
                                               model=queryParams[0],
                                               batch_size=queryParams[1],
                                               random_state=2317 \
                                               if fixRandomState else None)

    elif queryStrategyID==5:
        queryStrategy = SemiSupervisedBatchQuery(train_dataset,
                                                 model=queryParams[0],
                                                 batch_size=queryParams[1],
                                                 random_state=3112 \
                                                 if fixRandomState else None)

    return queryStrategy




"""Created by CRP #4 team between 20-27 August 2017.

Implementation of batch mode queries.
Uses the structure from libact - https://github.com/ntucllab/libact

- BatchQuery
    Batch query class
- RandomBatchQuery
    Random selection of batchSize objects
- LeastCertainBatchQuery
    Chooses the n objects with highest uncertainties
- SemiSupervisedBatchQuery
    Chooses the elements of the batch as detailed in
    XXX et al., 2017
"""

import libact
import copy
from libact.base.interfaces import QueryStrategy
from libact.base.dataset import Dataset
from libact.query_strategies import UncertaintySampling
from libact.utils import seed_random_state
from libact.base.interfaces import ContinuousModel, ProbabilisticModel

import numpy as np

class BatchQuery(QueryStrategy):

    def __init__(self, *args, **kwargs):
        super(BatchQuery, self).__init__(*args, **kwargs)
        
        random_state = kwargs.pop('random_state', None)
        self.random_state_ = seed_random_state(random_state)

        batch_size = kwargs.pop('batch_size', None)
        if (batch_size is None):
            raise TypeError(
                "__init__() missing required keyword-only argument: 'batch_size'"
            )
        
        self.batch_size_=batch_size
        
    def update(self, entry_id, label):
        pass

    def update_batch(self, entry_ids, labels):
        
        if len(entry_ids)==len(labels):
            
            for i in range(len(entry_ids)):
                self.dataset.update(entry_ids[i], labels[i])
            
        else:
            raise ValueError(
                "Id and label numbers not matching in update_batch"
            )
            
class RandomBatchQuery(BatchQuery):

    def __init__(self, *args, **kwargs):
        super(RandomBatchQuery, self).__init__(*args, **kwargs)
    
            
    def make_query(self):    
        
        dataset = self.dataset
        unlabeled_entry_ids, _ = zip(*dataset.get_unlabeled_entries())

        indx = self.random_state_.choice(len(unlabeled_entry_ids),
                                    size=self.batch_size_,replace=False)
        query_ids=[unlabeled_entry_ids[item] for item in indx]
        
        return query_ids

        
class LeastCertainBatchQuery(BatchQuery):

    def __init__(self, *args, **kwargs):
        super(LeastCertainBatchQuery, self).__init__(*args, **kwargs)
    
        self.model = kwargs.pop('model', None)
        if self.model is None:
            raise TypeError(
                "__init__() missing required keyword-only argument: 'model'"
            )
            
        if not isinstance(self.model, ContinuousModel) and \
                not isinstance(self.model, ProbabilisticModel):
            raise TypeError(
                "model has to be a ContinuousModel or ProbabilisticModel"
            )
        
        self.model.train(self.dataset)
            
    def update_batch(self, entry_ids, labels):
        super(LeastCertainBatchQuery, self).update_batch(entry_ids, labels)
        
        self.model.train(self.dataset)
        
    def make_query(self):
    
        unlabeled_entry_ids, unlabeled_entry_features = zip(*self.dataset.get_unlabeled_entries())
    
        probs = self.model.predict_proba(unlabeled_entry_features)

        indx = np.argsort(np.abs(probs[:,0] - 0.5))[:self.batch_size_]
        queryIds = [unlabeled_entry_ids[item] for item in indx]
        
        return queryIds

        
class SemiSupervisedBatchQuery(BatchQuery):

    def __init__(self, *args, **kwargs):
        super(SemiSupervisedBatchQuery, self).__init__(*args, **kwargs)
    
        self.model = kwargs.pop('model', None)
        if self.model is None:
            raise TypeError(
                "__init__() missing required keyword-only argument: 'model'"
            )
            
        if not isinstance(self.model, ContinuousModel) and \
                not isinstance(self.model, ProbabilisticModel):
            raise TypeError(
                "model has to be a ContinuousModel or ProbabilisticModel"
            )
        
        self.model.train(self.dataset)    

    def update_batch(self, entry_ids, labels):
        super(SemiSupervisedBatchQuery, self).update_batch(entry_ids,
                                                            labels)
        
        self.model.train(self.dataset)    
    
    def make_query(self):
        
        tempDataset = copy.deepcopy(self.dataset)    
        tempModel = copy.deepcopy(self.model)

        queryStrat = UncertaintySampling(tempDataset,
                                         model=tempModel) #Model is fit here
        queryIds = []    
        
        for j in range(self.batch_size_):
            queryId = queryStrat.make_query() #Model is also fit here
            queryIds.append(queryId)
            
            features = tempDataset.get_entries()[queryId][0]
            
            probs = tempModel.predict_proba(features.reshape(1, -1))
            
            # hard coded flag for positive answer - need to improve
            if self.random_state_.rand() < probs[0][0]:
                label = 0
            else:
                label = 1
            
            tempDataset.update(queryId,label)
            
            # tempModel.train(tempDataset) #This is not needed, 
            # since the make_query of UncertaintySampling fits
            
        return queryIds



"""Created by the CRP #4 team between 20-27 Aug 2017.

Functions to deal with the naming conventions on post-SNPCC data.

- snid2filename
    Convert snid to complete raw data file path.
"""

import os

def snid2filename(snid,data_path='.'): 
    """
    Convert snid to complete raw data file path.

    input: snid, str or int - supernova identification
           data_path, str - path to raw data folder

    output: str, complete path to raw data file.
    """

    # sanatize path
    sn_file = "DES_SN%s.DAT" % str(snid).zfill(6)

    return os.path.join(data_path,sn_file)



import numpy as np
import scipy.interpolate as interp 

def karpenka(time, A, B, t0, t1, t_fall, t_rise):
    """
    Double peaked exponential light curve function of Karpenka et al. (2013).
    Setting B to zero is the functional equivalent of Bazin (2011)

    :param time: the time in MJD
    :param A: normalization of first pulse
    :param B: normaliztion of second pulse
    :param t0: zero time of the pulse
    :param t1: zero time of the second pulse
    :param t_fall: decay tiem constant
    :param t_rise: rise time constant
    """
    t1m = time-t1
    t0m = time-t0

    return A*(1+ (B *t1m*t1m) )* np.exp(-t0m/t_fall )/ (1+ np.exp(-t0m/t_rise))

def sanders(time, Yb, Mp, alpha,  beta1,  beta2, betadN, betadC, t0,t1,tp,t2,td):
    """
    Type IIb/c supernova function from Sanders et al. (2014). These times are defined as durations
    not points.

    :param time: time in MJD
    :param Yb: background level
    :param alpha: first power law index
    :param beta1: second decay constant
    :param beta2: third decay constant
    :param betadC: forth decay constant
    :param betadN: fifth decay constant
    :param t0: start time of pulse
    :param t1: duration of rise
    :param tp: duration to peak
    :param t2: duration to decay
    :param td: duration to end

    """
    # normalize the constants
    # note: the paper has this wrong
    
    
    t = time - t0
    
    M1 = Mp/np.exp(beta1*tp)
    M2 = Mp/(np.exp(beta2*t2))
    Md = M2/np.exp(betadN*td)

    # select the time segments
    
    idx1=  (t>=0.) & (t<t1)
    idx2 = (t1<=t) & (t<t1+tp)
    idx3 = (t1+tp<=t) & (t<t1+tp+t2)
    idx4 = (t1+tp+t2<=t) & (t<t1+tp+t2+td)
    idx5 = t >= t1+tp+t2+td
    
    # calculate the power laws
    
    arg1 = M1*np.power(t[idx1]/t1,alpha)
    arg2 = M1*np.exp(beta1*(t[idx2]-t1))
    arg3 = Mp * np.exp(-beta2*(t[idx3]-(tp+t1)))
    arg4 = M2 * np.exp(-betadN*(t[idx4]-(t2+tp+t1)))
    arg5 = Md * np.exp(-betadC * (t[idx5] - (td+t2+tp+t1)))

    # now create an array that is just background
    
    out = np.zeros_like(t) + Yb

    # set the segments
    

    out[idx1] += arg1
    out[idx2] += arg2
    out[idx3] += arg3
    out[idx4] += arg4
    out[idx5] += arg5
    
  
    return out


def newling(time, A, phi, psi, k, sigma):
    """
    SN lightcurve function from Newling et al. (2011). 
    
    :param time: time in MJD
    :param A: Normalization
    :param phi: start time of pulse
    :param psi: tail value of pulse
    :param k: temporal streching
    :param sigma: temporal stretching
    """
    
    # calculate the spline function
    
    # create an empty array

    out = np.zeros_like(time)
    idx = time>=phi

    
    tail = np.zeros_like(time)

    # compute tau
    tau = k*sigma+phi
    
    # figure out where on the x-axis we compute things
    idx_spline = (phi<time) & (time<tau)
    idx_end = tau <= time
    
    # add the constants on
    tail[idx_end] = psi

    if (~idx_spline).sum()>=2:
        # create a spline in the regions between phi and tau    
        spline = interp.CubicSpline(time[~idx_spline],tail[~idx_spline])

        tail = spline(time)
   
    
    
    # now get the rest
    arg = (time[idx]-phi)/sigma
    
    out[idx] = A * np.power(arg,k) * np.exp(-arg) *np.power(k,-k) * np.exp(k)

    return out+tail


def bazin(time, A,t0,t_fall,t_rise):

    return karpenka(time,A,0.,t0,0.,t_fall,t_rise)









"""Created by Emille E. O. Ishida in 6 Sep 2017.

Configuration function to choose between different active learning settings.

- ConfigureStrategy
    Build dictionary with possible configurations.
"""

from actsnclass.actConfig.initialSetup import initialDataSetup
from actsnclass.actConfig.initialSetup import initialModelSetup,initialQuerySetup

def ConfigureStrategy(chosenInput):
    """
    Build dictionary with possible configurations. 

    input: chosenInput, dict - user choices
           keywords must include:
               choice, str - one of the following
                 'rand_RF': random sampling with random forest
                 'unc_RF ': uncertainty sampling with random forest
                 'qbc_noble': query by committee, a'la Noble
                 'qbc_robSmall' : query by committee, a'la Robert, 
                                                      small committee
                 'qbc_robBig': query by committee, a'la Robert, 
                                                   big committee
                 'batch_random': batch random sampling, a'la Jim
                 'batch_nlunc': batch N-least uncertain, a'la Jim
                 'batch_semi': batch semi-supervised, a'la Jim
               useOnlyFirstHalf, boolean - rather to apply epoch cuts
               useFixedRandomState, boolean - rather to fix random seed
               validationDataProportion, float - fraction of data separated 
                                                 for validation
               batchSize, int - number of SNe query per batch mode
               queryQuota, int - total number of queries
               mainTreeSize, int - number of trees in Random Forest
               treesInRFCommittee, int - number of RF class in the committee

    output: final_config, dict - parameters for chosen configuration
            keywords are:
                runSetupID, int - numerical identification of configuration
                output_name, str - root of output file names
                modelID, int - model identification
                modelParams, tuple - parameters of the model
                queryStrategyID, int - query strategy identification
                queryParams, tuple or None - parameters of query strategy
                quota, int - number of iterations calling the AL routine 
    """

    config = {}
    config['random_RF'] = {}
    config['random_RF']['runSetupID'] = 0
    config['random_RF']['output_name'] = 'random_RF'
    config['random_RF']['modelID'] = 0
    config['random_RF']['modelParams'] = (chosenInput['mainTreeSize'],)
    config['random_RF']['queryStrategyID'] = 0
    config['random_RF']['queryParams'] = None
    config['random_RF']['quota'] = chosenInput['queryQuota']

    config['unc_RF'] = {}
    config['unc_RF']['runSetupID'] = 1
    config['unc_RF']['output_name'] = 'unc_RF'
    config['unc_RF']['modelID'] = 0
    config['unc_RF']['modelParams'] = (chosenInput['mainTreeSize'],)
    config['unc_RF']['queryStrategyID'] = 1
    config['unc_RF']['quota'] = chosenInput['queryQuota']

    config['qbc_noble'] = {}
    config['qbc_noble']['runSetupID'] = 2
    config['qbc_noble']['output_name'] = 'qbc_noble'
    config['qbc_noble']['modelID'] = 1
    config['qbc_noble']['modelParams'] = (chosenInput['mainTreeSize'], \
                                          chosenInput['treesInRFCommittee'])
    config['qbc_noble']['queryStrategyID'] = 2
    config['qbc_noble']['quota'] = chosenInput['queryQuota']

    config['qbc_robSmall'] = {}
    config['qbc_robSmall']['runSetupID'] = 3
    config['qbc_robSmall']['output_name'] = 'qbc_robSmall'
    config['qbc_robSmall']['modelID'] = 2
    config['qbc_robSmall']['modelParams'] = (None,)
    config['qbc_robSmall']['queryStrategyID'] = 2
    config['qbc_robSmall']['quota'] = chosenInput['queryQuota']

    config['qbc_robBig'] = {}
    config['qbc_robBig']['runSetupID'] = 4
    config['qbc_robBig']['output_name'] = 'qbc_robBig'
    config['qbc_robBig']['modelID'] = 3
    config['qbc_robBig']['modelParams'] = (None,)
    config['qbc_robBig']['queryStrategyID'] = 2
    config['qbc_robBig']['quota'] = chosenInput['queryQuota']

    config['batch_random'] = {}
    config['batch_random']['runSetupID'] = 5
    config['batch_random']['output_name'] = 'batch_random'
    config['batch_random']['modelID'] = 0
    config['batch_random']['modelParams'] = (chosenInput['mainTreeSize'],)
    config['batch_random']['queryStrategyID'] = 3
    config['batch_random']['queryParams'] = (chosenInput['batchSize'],)
    config['batch_random']['quota'] = \
                      chosenInput['queryQuota'] // chosenInput['batchSize']

    config['batch_nlunc'] = {}
    config['batch_nlunc']['runSetupID'] = 6
    config['batch_nlunc']['output_name'] = 'batch_nlunc'
    config['batch_nlunc']['modelID'] = 0
    config['batch_nlunc']['modelParams'] = (chosenInput['mainTreeSize'],)
    config['batch_nlunc']['queryStrategyID'] = 4
    config['batch_nlunc']['quota'] = \
                       chosenInput['queryQuota']//chosenInput['batchSize']

    config['batch_semi'] = {}
    config['batch_semi']['runSetupID'] = 7
    config['batch_semi']['output_name'] = 'batch_semi'
    config['batch_semi']['modelID'] = 0
    config['batch_semi']['modelParams'] = (chosenInput['mainTreeSize'],)
    config['batch_semi']['queryStrategyID'] = 5
    config['batch_semi']['quota'] = \
                       chosenInput['queryQuota']//chosenInput['batchSize']

    # Create list of models to be trained
    config['random_RF']['mainModelList'] = \
                initialModelSetup(config['random_RF']['modelID'],
                                  config['random_RF']['modelParams'],
                                  chosenInput['useFixedRandomState'])
    config['unc_RF']['mainModelList'] = \
                initialModelSetup(config['unc_RF']['modelID'],
                                  config['unc_RF']['modelParams'],
                                  chosenInput['useFixedRandomState'])
    config['qbc_noble']['mainModelList'] = \
                initialModelSetup(config['qbc_noble']['modelID'],
                                  config['qbc_noble']['modelParams'],
                                  chosenInput['useFixedRandomState'])
    config['qbc_robSmall']['ModelList'] = \
          initialModelSetup(config['qbc_robSmall']['modelID'], 
                            config['qbc_robSmall']['modelParams'],
                            chosenInput['useFixedRandomState'])
    config['qbc_robSmall']['mainModelList'] = \
          initialModelSetup(0, (chosenInput['mainTreeSize'],),
                                chosenInput['useFixedRandomState'])
    config['qbc_robBig']['ModelList'] = \
          initialModelSetup(config['qbc_robBig']['modelID'], 
                            config['qbc_robBig']['modelParams'],
                            chosenInput['useFixedRandomState'])
    config['qbc_robBig']['mainModelList'] = \
          initialModelSetup(0, (chosenInput['mainTreeSize'],), 
                           chosenInput['useFixedRandomState'])
    config['batch_random']['mainModelList'] = \
                initialModelSetup(config['batch_random']['modelID'],
                                  config['batch_random']['modelParams'],
                                  chosenInput['useFixedRandomState'])
    config['batch_nlunc']['mainModelList'] = \
                initialModelSetup(config['batch_nlunc']['modelID'],
                                  config['batch_nlunc']['modelParams'],
                                  chosenInput['useFixedRandomState'])
    config['batch_semi']['mainModelList'] = \
                initialModelSetup(config['batch_semi']['modelID'],
                                  config['batch_semi']['modelParams'],
                                  chosenInput['useFixedRandomState'])

    # determine query parameters
    config['unc_RF']['queryParams'] = (config['unc_RF']['mainModelList'][0],)
    config['qbc_noble']['queryParams'] = (config['qbc_noble']['mainModelList'],)
    config['qbc_robSmall']['queryParams'] = (config['qbc_robSmall']['ModelList'],)
    config['qbc_robBig']['queryParams'] = (config['qbc_robBig']['ModelList'],)
    config['batch_nlunc']['queryParams'] = \
        (config['batch_nlunc']['mainModelList'][0], chosenInput['batchSize'])
    config['batch_semi']['queryParams'] = \
        (config['batch_semi']['mainModelList'][0], chosenInput['batchSize'])

    # Select choice
    final_config = config[chosenInput['choice']]

    if chosenInput['useOnlyFirstHalf']:
        final_config['queryFile'] = chosenInput['outputFolder'] + 'query_' + \
                                                   final_config['output_name'] + '_halfLC.csv'
        final_config['diagnosticFile'] = chosenInput['outputFolder'] + \
                                   'diagnostic_' + final_config['output_name'] + '_halfLC.csv'
    else:
        final_config['queryFile'] = chosenInput['outputFolder'] + 'query_' + \
                                               final_config['output_name'] + '_fullLC_v2.csv'
        final_config['diagnosticFile'] = chosenInput['outputFolder'] + \
                               'diagnostic_' + final_config['output_name'] + '_fullLC_v2.csv'

    return final_config



"""Created by CRP #4 Team between 20-27 Aug 2017
"""

import pandas as pd
import numpy as np

class MLResult(object):
    
    _required_keys = ["fom", "efficiency", "purity", "accuracy"]
    
    def __init__(self, snids, fom, efficiency, purity, accuracy, **kwargs):
    
        table = {}
        
        n_entries = len(snids)
        # Create number of queries vector
        indx = range(0, n_entries)
        
        #table['snids'] = snids
        table['fom'] = fom
        table['efficiency'] = efficiency
        table['purity'] = purity
        table['accuracy'] = accuracy

        diag_rows = np.vstack((indx, snids, accuracy, efficiency, purity, fom))
        diag_columns = diag_rows.transpose()
        
        for key, value in kwargs.items():
            
            table[key] = value
            
        for key in MLResult._required_keys:
            
            assert len(table[key]) == n_entries, '%s has the wrong length'%key
            
        for key in kwargs.keys():
            
            assert len(table[key]) == n_entries, '%s has the wrong length'%key

        self._table = pd.DataFrame(diag_columns, columns=['nqueries', 'snids', 'accuracy', 
                                                'eff', 'pur', 'fom'])
        self._table['nqueries'] = self._table['nqueries'].astype(int)      
        
    def save(self,file_name):
       
        self._table.to_csv(file_name, 
                           index=False, header=['nqueries', 'snids', 'accuracy', 'eff',
                                        'pur', 'fom'])
        
    @property
    def results(self):

        return self._table

        
    @classmethod
    def from_save(file_name):
        return pd.DataFrame.from_csv(file_name, header=True)
            
            
            
            
        
        
        
    



import os
from actsnclass.data_functions.read_SNANA import read_snana_lc
import numpy as np
import matplotlib.pyplot as plt
import copy
from scipy.stats import norm
from mpl_toolkits.mplot3d import Axes3D
import collections

try:
    from threeML import *

    no_threeml = False
    from threeML.io.file_utils import file_existing_and_readable

except:

    no_threeml = True

try:
    import pygmo
    no_pygmo = False
except ImportError:

    no_pygmo = True

no_threeml = False
no_pygmo = False

class SNObservation(object):
    def __init__(self, data_file, snr_table=None):

        header, flux = read_snana_lc(data_file, flux_as_DataFrame=True)
        self._sample = header['sample']

        self._pkmjd = header['pkmjd']

        self._pkmag = {}
        for band, val in zip(['g', 'i', 'r', 'z'], header['pkmag']):
            self._pkmag[band] = val

        self._type = header['type'][0]
        if self._type != 0:
            self._is_1a = False
        else:
            self._is_1a = True
        self._snid = header['snid']
        self._z = header['z']
        self._model = header['model']
        self._bands = collections.OrderedDict()

        if snr_table is not None:

            head, tail = os.path.split(data_file)

            use_band = snr_table[snr_table['sn_file'] == tail][0]

        else:

            use_band = {'g': True, 'r': True, 'i': True, 'z': True}

        for band, flux_table in flux.iteritems():

            if use_band[band]:

                sn_band = SNBand(flux_table, band, self)
                self._bands[band] = sn_band

                setattr(self, band, sn_band)

    def __repr__(self):

        s = ''

        s += 'snid: %d\n' % self._snid
        s += 'sample: %s\n' % self._sample
        s += 'is Ia: %s\n' % self._is_1a
        s += 'type: %s\n' % self._type
        s += 'bands: %s' % ' ,'.join(self._bands)
        return s

    @property
    def sample(self):
        return self._sample

    @property
    def pkmag(self):

        return self._pkmag

    @property
    def type(self):

        return self._type

    @property
    def is_1a(self):
        return self._is_1a

    @property
    def redshift(self):
        return self._z

    @property
    def snid(self):
        return self._snid

    @property
    def pkmjd(self):

        return self._pkmjd

    @property
    def model(self):

        return self._model

    def plot(self):
        """
        plot the light curves
        """

        fig, ax = plt.subplots()

        for band in self._bands.itervalues():

            _ = band.plot(ax=ax)

        ax.legend()

        return fig

    def fit(self,
            function,
            with_global_optimization=True,
            bands=None,
            quiet=False):

        if bands is None:
            bands = self._bands.keys()

        for band in bands:

            new_func = copy.deepcopy(function)

            self._bands[band].fit(new_func,
                                  with_global_optimization,
                                  quiet=quiet)

    def fit_multinest(self,
                      function,
                      bands=None,
                      quiet=False,
                      resume=False,
                      chain_name='',
                      check_fit=True,
                      destination=None,
                      info=None):

        if bands is None:
            bands = self._bands.keys()

        for band in bands:

            dontgo = True
            if check_fit:

                dontgo = self._bands[band].check_if_fit_exists(destination,
                                                               info)

            else:

                dontgo = False

            if not dontgo:

                new_func = copy.deepcopy(function)

                self._bands[band].fit_multinest(
                    new_func,
                    quiet=quiet,
                    resume=resume,
                    chain_name="%s_%s-" % (chain_name, band))

    def plot_fit(self):

        fig, ax = plt.subplots()

        for v in self._bands.itervalues():

            try:

                _ = v.plot_fit(ax)

            except:

                pass

        return fig

    def save_fit(self, destination='.', fit_info='', overwrite=False):

        for v in self._bands.itervalues():

            v.save_fit(destination, fit_info, overwrite)

    @property
    def bands(self):

        return self._bands

    @property
    def peak_flux(self):

        peak_flux = {}

        for k, v in self._bands.iteritems():

            peak_flux[k] = v.peak_flux

        return peak_flux

    @property
    def peak_time(self):

        peak_time = {}

        for k, v in self._bands.iteritems():

            peak_time[k] = v.peak_time

        return peak_time


class SNBand(object):

    _color_wheel = {
        'r': '#00ff7b',
        'g': '#00e1ff',
        'i': '#ff0000',
        'z': '#FCFF31'
        #        'z': '#ff7b00'
    }

    def __init__(self, flux_table, band, observation):

        self._observation = observation
        self._snid = observation.snid

        self._band = band

        err_idx = np.isfinite(np.array(flux_table['FLUXCALERR']))

        self._error = np.array(flux_table['FLUXCALERR'])[err_idx]

        self._mjd = np.array(flux_table['MJD'])[err_idx]
        self._flux = np.array(flux_table['FLUXCAL'])[err_idx]
        self._error = np.array(flux_table['FLUXCALERR'])[err_idx]
        self._snr = np.array(flux_table['SNR'])[err_idx]

        self._zero_time = self._mjd.min()

        self._idx_max = np.argmax(self._flux)

        self._normalized_time = self._mjd - self._zero_time

        self._jl = None

    @property
    def mjd(self):
        return self._mjd

    @property
    def normalized_time(self):

        return self._normalized_time

    @property
    def flux(self):
        return self._flux

    @property
    def error(self):
        return self._error

    @property
    def peak_time(self):

        return self._normalized_time[self._idx_max]

    @property
    def peak_flux(self):

        return self._flux[self._idx_max]

    @property
    def zero_time(self):

        return self._zero_time

    def plot(self, ax=None, norm_time=False):

        if ax is None:

            fig, ax = plt.subplots()

        else:

            fig = ax.get_figure()

        if norm_time:
            time = self._normalized_time

        else:

            time = self._mjd

        ax.errorbar(
            time,
            self._flux,
            yerr=self._error,
            color=self._color_wheel[self._band],
            fmt='o',
            label=self._band)

        return fig

    def fit_multinest(self, function, resume=False, chain_name='',
                      quiet=False):

        self._ps = PointSource(self._band, 0, 0, spectral_shape=function)
        self._model = Model(self._ps)

        self._xyl = XYLike('band_%s' % (self._band), self._normalized_time,
                           self._flux, self._error)
        self._jl = BayesianAnalysis(self._model, DataList(self._xyl))

        self._jl.sample_multinest(
            n_live_points=300,
            quiet=quiet,
            chain_name=chain_name,
            resume=resume,
            importance_nested_sampling=False)

    def fit(self, function, with_global_optimization=True, quiet=False):

        if no_threeml:
            raise RuntimeError('No 3ML!')

        self._ps = PointSource(self._band, 0, 0, spectral_shape=function)
        self._model = Model(self._ps)

        self._xyl = XYLike('band_%s' % (self._band), self._normalized_time,
                           self._flux, self._error)
        self._jl = JointLikelihood(self._model, DataList(self._xyl))

        if with_global_optimization:
            pagmo_minimizer = GlobalMinimization("pagmo")

            my_algorithm = pygmo.algorithm(pygmo.de(gen=25))

            # Create an instance of a local minimizer
            local_minimizer = LocalMinimization("ROOT")

            # Setup the global minimization
            pagmo_minimizer.setup(
                second_minimization=local_minimizer,
                algorithm=my_algorithm,
                islands=10,
                population_size=15,
                evolution_cycles=5)

            self._jl.set_minimizer(pagmo_minimizer)

        else:

            self._jl.set_minimizer('ROOT')

        try:

            _ = self._jl.fit(quiet=quiet)

        except:

            print('Fit FAILED')

            self._jl = None

    def check_if_fit_exists(self, destination='.', save_info=''):

        file_name = "%s/%s_%s_%s.fits" % (destination, self._band, self._snid,
                                          save_info)

        return file_existing_and_readable(file_name)

    def save_fit(self, destination='.', save_info='', overwrite=True):

        if self._jl is not None:

            file_name = "%s/%s_%s_%s.fits" % (destination, self._band,
                                              self._snid, save_info)

            self._jl.results.write_to(file_name, overwrite=overwrite)

    def plot_fit(self, ax=None):

        if self._jl is not None:

            return self._xyl.plot(
                x_label='time',
                y_label='flux',
                subplot=ax,
                model_color='k',
                data_color=self._color_wheel[self._band])

        else:
            print('T11111')


data_path = os.path.join('..', 'data', 'SIMGEN_PUBLIC_DES')


class FittedObservation(object):
    def __init__(self, *fitted_bands, **kwargs):
        """
        :param fitted_bands: filenames for each band of this fit
        :param data_dir: the location of the DES data files
        """

        self._bands = {}
        tmp_snids = []
        tmp_functions = []

        for f in fitted_bands:

            filename = f.split('/')[-1]

            band, snid, ext = filename.split('_')

            function, _ = ext.split('.')

            ar = load_analysis_results(f)

            self._bands[band] = ar

            tmp_snids.append(snid)
            tmp_functions.append(function)

            setattr(self, band, ar)

        assert len(np.unique(tmp_snids)) == 1
        assert len(np.unique(tmp_functions)) == 1

        if 'data_dir' in kwargs:

            data_dir = kwargs.pop('data_dir')

            snid = tmp_snids[0]

            snid_str = snid.zfill(6)

            data_file = os.path.join(data_dir, 'DES_SN%s.DAT' % snid_str)
            self._snobs = SNObservation(data_file)

        else:

            self._snobs = None

    def extract_credible_regions(self, time_range, band):

        assert band in self._bands.keys()

        ps = self._bands[band].optimized_model.point_sources[band]

        parameters = ps.spectrum.main.shape.parameters
        test_model = ps.spectrum.main.shape
        parameter_names = [
            par.name for par in ps.spectrum.main.shape.parameters.values()
        ]

        arguments = {}

        # because we might be using composite functions,
        # we have to keep track of parameter names in a non-elegant way
        for par, name in zip(parameters.values(), parameter_names):

            if par.free:

                this_variate = self._bands[band].get_variates(par.path)

                # Do not use more than 1000 values (would make computation too slow for nothing)

                if len(this_variate) > 1000:
                    this_variate = np.random.choice(this_variate, size=1000)

                    arguments[name] = this_variate

            else:

                # use the fixed value rather than a variate

                arguments[name] = par.value

                # create the propagtor

        propagated_function = self._bands[band].propagate(test_model.evaluate,
                                                          **arguments)

        return [propagated_function(t) for t in time_range]

    def posterior_plot(self, time_range, thin=10, with_data=True):

        trace_colors = {
            'r': '#076d00',
            'g': '#008b9b',
            'i': '#a50018',
            'z': '#9b4800'
        }

        fig, ax = plt.subplots()

        for band in self._bands:

            # plot the data
            if with_data:
                self._snobs.bands[band].plot(ax, norm_time=True)

            # now build the samples range
            variates_in_time = self.extract_credible_regions(time_range, band)
            samples_in_time = np.array(
                [v.samples[::thin].tolist() for v in variates_in_time]).T

            for sample in samples_in_time:

                if with_data:

                    ax.plot(
                        time_range, sample, color=trace_colors[band], alpha=.1)

                else:
                    ax.plot(
                        time_range,
                        sample,
                        color=SNBand._color_wheel[band],
                        alpha=.1)

        ax.legend()
        return fig

    def posterior_plot_3D(self,
                          time_range,
                          thin=10,
                          with_data=False,
                          with_cloud=True,
                          n_cloud_points=20,
                          point_size=20,
                          point_color='k',axis_color='k',bkg_color=None):

        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        for k, band in enumerate(self._bands):

            # now build the samples range
            variates_in_time = self.extract_credible_regions(time_range, band)
            samples_in_time = np.array(
                [v.samples[::thin].tolist() for v in variates_in_time]).T

            if with_data:

                xs = self._snobs.bands[band].normalized_time
                ys = self._snobs.bands[band].flux
                err = self._snobs.bands[band].error
                if not with_cloud:
                    ax.scatter(
                        [k + 1] * len(xs), xs, ys, c=point_color, s=point_size)
                    for x,y,e in zip(xs,ys,err):

                        ax.plot([k+1,k+1],[x,x],[y-e,y+e], lw=1, color=point_color)
                    
                else:

                    for x, y, e in zip(xs, ys, err):

                        new_k = norm.rvs(loc=[k + 1],
                                         scale=.1,
                                         size=n_cloud_points)
                        new_x = norm.rvs(loc=x, scale=.1, size=n_cloud_points)
                        new_y = norm.rvs(loc=y, scale=e, size=n_cloud_points)

                        for a, b, c in zip(new_k, new_x, new_y):
                            alp1 = norm.pdf(c, loc=y, scale=e)
                            
                            ax.scatter(
                                a,
                                b,
                                c,
                                c=point_color,
                                s=point_size,
                                alpha=alp1)

            for sample in samples_in_time:

                xx = norm.rvs(loc=k + 1, scale=.1)

                alpha = .05 * xx / float(k + 1)  #norm.rvs(loc=.2,scale=.1)

                ax.plot(
                    [xx] * len(time_range),
                    time_range,
                    sample,
                    lw=2,
                    color=SNBand._color_wheel[band],
                    alpha=alpha)

        ax.set_ylabel('Time')
        ax.set_zlabel('Flux')

        ax.grid(False)

        ax.xaxis.pane.set_edgecolor(axis_color)
        ax.yaxis.pane.set_edgecolor(axis_color)
        if bkg_color is not None:
            ax.xaxis.pane.set_color(bkg_color)
            ax.yaxis.pane.set_color(bkg_color)
            ax.zaxis.pane.set_color(bkg_color)

        else:
            ax.xaxis.pane.fill = False
            ax.yaxis.pane.fill = False
            ax.zaxis.pane.fill = False

        ax.set_zlim(0.)
        ax.w_zaxis.line.set_color((1.0, 1.0, 1.0, 0.0))
        ax.set_zticks([])

        plt.xticks(np.arange(len(self._bands)) + 1, self._bands)

    @property
    def sn_obs(self):

        return self._snobs

    def plot(self):

        pass
